"""
ncOrtho - Targeted ortholog search for miRNAs
Copyright (C) 2021 Felix Langschied

ncOrtho is a free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

ncOrtho is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ncOrtho.  If not, see <http://www.gnu.org/licenses/>.
"""

import pyfaidx

class GenomeParser(object):
    
    # genpath: path to the genome file to extract from
    # hitlist: list of hits to extract the sequence for
    def __init__(self, genpath, hitlist):
        self.genpath = genpath
        self.hitlist = hitlist
        self.gene_dict = self.parse_genome()
    
    # read in the genome
    def parse_genome(self,):
        genome = pyfaidx.Fasta(self.genpath)
        return genome
    
    # extract the sequences for each significant hit from cmsearch
    def extract_sequences(self,):
        # dictionary to store the sequences
        seq_dict = {}
        for hit in self.hitlist:
            # hit on sense strand, sequence can be extracted directly
            if hit[4] == '+':
                seq = self.gene_dict[hit[1]][int(hit[2])-1:int(hit[3])].seq
            # hit on antisense strand, sequence has to be extracted
            # as reverse complement
            # note that cmsearch switches start and end positions in case
            # of antisense strand hits
            elif hit[4] == '-':
                seq = (
                    self.gene_dict[hit[1]][int(hit[2])-1:int(hit[3])]
                    .reverse.complement.seq
                )
            seq_dict[hit[0]] = seq
        return seq_dict



# Construct and calibrate a covariance model for a given ncRNA core set
# alignment in Stockholm format.
# The calibration step is computationally very expensive and should be
# performed on multiple CPU cores.
# TODO: Include license, author etc.

import argparse
import multiprocessing as mp
import os
import subprocess as sp
import sys


class CmConstructor(object):
    
    def __init__(self, alignment, outpath, name, cpu):
        self.alignment = alignment
        self.outpath = outpath
        self.name = name
        self.cpu = cpu
        self.model = '{0}/{1}.cm'.format(outpath, name)
    
    def construct(self):
        print('# Constructing covariance model for {}.'.format(self.name))
        construct_command = (
            'cmbuild -n {0} {1}/{0}.cm {2}'
            .format(self.name, self.outpath, self.alignment)
        )
        sp.call(construct_command, shell=True)
        print('# Finished covariance model construction.')
    
    def calibrate(self):
        print('# Calibrating covariance model for {}.'.format(self.name))
        calibrate_command = (
            'cmcalibrate --cpu {0} {1}'.format(self.cpu, self.model)
        )
        sp.call(calibrate_command, shell=True)
        print('# Finished covariance model calibration.')


def main():

    # Parse command-line arguments.
    parser = argparse.ArgumentParser(
        prog='python createcm.py', description='covariance model construction'
    )
    # "cpu", use maximum number of available CPUs unless specified otherwise.
    parser.add_argument(
        '-c', '--cpu', metavar='int', type=int,
        help='number of CPU cores to use', nargs='?',
        const=mp.cpu_count(), default=mp.cpu_count()
    )
    # Path to the desired output folder.
    parser.add_argument(
        '-o', '--output', metavar='<path>', type=str,
        help='path to the output folder'
    )
    # Path to the input alignment
    parser.add_argument(
        '-a', '--alignment', metavar='<.sto>', type=str,
        help='path to input alignment'
    )

    # Show help when no arguments are added.
    if len(sys.argv) == 1:
        parser.print_help()
        sys.exit(1)
    else:
        args = parser.parse_args()
    
    # Check if computer provides the desired number of cores.
    available_cpu = mp.cpu_count()
    if args.cpu > available_cpu:
        print(
            '# Error: The provided number of CPU cores is higher than the '
            'number available on this system. Exiting...'
        )
        sys.exit(1)
    else:
        cpu = args.cpu

    # Check if alignment file exists.
    if not os.path.isfile(args.alignment):
        print(
            '# Error: The provided path to the alignment file appears to be '
            'invalid. Exiting...'
        )
        sys.exit(1)
    else:
        alignment = args.alignment

    output = args.output
    alignment = args.alignment
    name = alignment.split('/')[-1].split('.')[0]

    # Check if the output folder exists.
    if not os.path.isdir(output):
        mkdir_cmd = 'mkdir -p {}'.format(output)
        sp.call(mkdir_cmd, shell=True)

    # Initiate covariance model construction and calibration.
    cmc = CmConstructor(alignment, output, name, cpu)
    # Construct the model.
    cmc.construct()
    # Calibrate the model.
    cmc.calibrate()


if __name__ == '__main__':
    main()



"""
ncOrtho - Targeted ortholog search for miRNAs
Copyright (C) 2021 Felix Langschied

ncOrtho is a free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

ncOrtho is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ncOrtho.  If not, see <http://www.gnu.org/licenses/>.
"""

# Create a core set of orthologs
# Find the corresponding syntenic regions in reference and core species
# Search for core orthologs by reciprocal BLAST search
# Create Stockholm structural alignment

# required:
# reference microRNA data (sequence, coordinates)
# reference taxon: genome, blastdb, gtf file with gene coordinates
# core set taxa: genome, gtf file, pairwise orthologs

import argparse
import multiprocessing as mp
import os
import pyfaidx
import subprocess as sp
import sys
import yaml


from ncOrtho.coreset.createcm import CmConstructor
from ncOrtho.coreset.core_reblast import blastsearch
from ncOrtho.coreset.coreset_utils import gff_parser
from ncOrtho.coreset.coreset_utils import gtf_parser
from ncOrtho.coreset.coreset_utils import ortho_search

# from createcm import CmConstructor
# from core_reblast import blastsearch
# from coreset_utils import gff_parser
# from coreset_utils import gtf_parser
# from coreset_utils import ortho_search


###############################################################################
def main():
    # Print header
    print('\n' + '#' * 43)
    print('###' + ' ' * 37 + '###')
    print('###   ncOrtho - core set construction   ###')
    print('###' + ' ' * 37 + '###')
    print('#' * 43 + '\n')

    # Parse command-line arguments
    # Define global variables
    parser = argparse.ArgumentParser(
        description=(
            'Build Covariance models of reference miRNAs from core set of orthologs.'
         )
    )
    parser._action_groups.pop()
    required = parser.add_argument_group('Required Arguments')
    optional = parser.add_argument_group('Optional Arguments')
    # input file path
    required.add_argument('-p', '--parameters', metavar='<path>', type=str, required=True,
                        help='Path to the parameters file in yaml format')
    # mirna data
    required.add_argument(
        '-n', '--ncrna', metavar='<path>', type=str, required=True,
        help='Path to tab separated file of reference miRNAs information '
    )
    # output folder
    required.add_argument(
        '-o', '--output', metavar='<path>', type=str, required=True,
        help='Path for the output folder'
    )
    # OPTIONAL VARIABLES
    # cpu, use maximum number of available cpus if not specified otherwise
    optional.add_argument(
        '--threads', metavar='int', type=int,
        help='Number of CPU cores to use (Default: All available)', nargs='?',
        const=mp.cpu_count(), default=mp.cpu_count()
    )
    # Maximum gene insertions
    optional.add_argument(
        '--mgi', metavar='int', type=int,
        help='Maximum number of gene insertions in the core species (Default: 3)', nargs='?',
        const=3, default=3
    )
    # use dust filter?
    optional.add_argument(
        '--dust', metavar='yes/no', type=str,
        help='Use BLASTn dust filter. '
             'Will not build models from repeat regions. (Default: yes)',
        nargs='?',
        const='yes', default='yes'
    )
    optional.add_argument(
        '--create_model', metavar='yes/no', type=str,
        help='set to "no" if you only want to create the alignment. (Default: yes)', nargs='?',
        const='yes', default='yes'
    )
    #
    optional.add_argument(
        '--idtype', metavar='str', type=str,
        help='Choose the ID in the reference gff file that is '
             'compared to the IDs in the pairwise orthologs file (default:ID) '
             'Options: ID, Name, GeneID, gene_id',
        nargs='?', const='ID=', default='ID'
    )

    ###############################################################################

    # Show help when no arguments are added.
    if len(sys.argv) == 1:
        parser.print_help()
        sys.exit(1)
    else:
        args = parser.parse_args()

    # Check if computer provides the desired number of cores.
    available_cpu = mp.cpu_count()
    if args.threads > available_cpu:
        print(
            '# Error: The provided number of CPU cores is higher than the '
            'number available on this system. Exiting...'
        )
        sys.exit(1)
    else:
        cpu = args.threads

    # parse mandatory arguments
    mirna_path = args.ncrna
    output = args.output

    param_file = args.parameters
    path_dict = {}  # {<name>: {'orthologs': <path>, 'genome': <path>, 'annotation': <path>}}
    with open(param_file, 'r') as param_handle:
        params = yaml.load_all(param_handle, Loader=yaml.FullLoader)
        for entry in params:
            in_type = entry.pop('type')
            if in_type == 'reference':
                ref_annot_path = entry['annotation']
                ref_genome = entry['genome']
            elif in_type == 'core':
                species = entry.pop('name')
                path_dict[species] = entry

    # parse optional arguments
    mgi = args.mgi
    dust = args.dust
    id_t = args.idtype
    create_model = args.create_model

    # check if files exist
    all_paths = []
    for sd in path_dict:
        all_paths.extend(list(path_dict[sd].values()))
    all_paths.extend([ref_annot_path, ref_genome])
    for cp in all_paths:
        if not os.path.isfile(cp):
            print(f'ERROR: {cp} does not exist')
            sys.exit()

    ###############################################################################
    ortho_dict = {}
    neighbor_dict = {}
    mirna_dict = {}

    print('# Reading pairwise orthologs')
    for taxon in path_dict:
        with open(path_dict[taxon]['orthologs']) as omafile:
            oma_lines = omafile.readlines()
            orthologs = {
                ref: core for (ref, core) in [
                    (line.split()[0], line.split()[1])
                    for line in oma_lines if not line.startswith('#')
                ]
            }
            ortho_dict[taxon] = orthologs
    print('Done')

    print('# Reading miRNA data')
    with open(mirna_path) as mirfile:
        mirnas = [
            line.split() for line in mirfile.readlines()
            if not line.startswith('#')
        ]
    print('Done')

    print('# Reading reference annotation')
    ft = ref_annot_path.split('.')[-1]
    if ft == 'gtf':
        ref_dict = gtf_parser(ref_annot_path)
    elif ft in ['gff3', 'gff']:
        ref_dict = gff_parser(ref_annot_path, id_t)
    else:
        print('ERROR: File type "{}" not valid as reference annotation'.format(ft))
        sys.exit()
    print('Done')

    # Determine the position of each miRNA and its neighboring gene(s)
    for mirna in mirnas:
        sys.stdout.flush()
        mirid = mirna[0]
        print('### {0} ###'.format(mirid))
        # Check if output folder exists or create it otherwise
        if not os.path.isdir('{}/{}'.format(output, mirid)):
            sp.call('mkdir -p {}/{}'.format(output, mirid), shell=True)
        ### Workaround for differing naming conventions in miRBase and Ensembl
        if 'chr' in mirna[1]:
            chromo = mirna[1].split('chr')[1]
        else:
            chromo = mirna[1]
        ##
        start = int(mirna[2])
        end = int(mirna[3])
        strand = mirna[4]

        ### find left neighbor or check if located inside gene
        ### chr_dict[contig][i] = (geneid, start, end, strand)
        ###############################################################################
        # case 1): there is no protein-coding gene on the same contig as the miRNA,
        # so there can be no neighbors (should only occur in highly fragmented
        # assemblies)
        if not chromo in ref_dict.keys():
            print(
                'WARNING: No protein-coding genes found on contig "{0}". '
                'Synteny around {1} cannot be established.\n'
                'Make sure that the contig identifiers of the miRNA input file '
                'match the ones in the reference annotation file'
                    .format(chromo, mirid)
            )
            continue

        # case 2): miRNA is located left of the first gene and hence has no left
        # neighbor, the first gene is therefore by default the right neighbor
        if end < int(ref_dict[chromo][1][1]):
            print(
                'There is no left neighbor of {0}, since it is located at the '
                'start of contig {1}.'.format(mirid, chromo)
            )
            print(
                '{0} is the right neighbor of {1}.'
                    .format(ref_dict[chromo][1][0], mirid)
            )
            continue

        # case 3): miRNA is located right to the last gene, so the last gene is the
        # left neighbor and there cannot be a right neighbor
        elif start > int(ref_dict[chromo][len(ref_dict[chromo])][2]):
            print(
                '{0} is the left neighbor of {1}.'
                    .format(ref_dict[chromo][len(ref_dict[chromo])][0], mirid)
            )
            print(
                'There is no right neighbor of {0}, since it is located at the'
                ' end of contig {1}.'.format(mirid, chromo)
            )
            continue

        # case 4): miRNA is located either between two genes or overlapping with (an
        # intron of) a gene, either on the same or the opposite strand
        ###############################################################################
        else:
            solved = False
            for i, gene in enumerate(ref_dict[chromo]):
                gene_data = ref_dict[chromo][gene]
                # case 4.1): miRNA inside gene
                if (
                        start >= gene_data[1]
                        and end <= gene_data[2]
                        and strand == gene_data[3]
                ):
                    solved = True
                    # c+=1
                    print(
                        '{0} is located inside the gene {1}.'
                            .format(mirid, gene_data[0])
                    )
                    ortho_hits = ortho_search(gene_data[0], ortho_dict)
                    for core_tax in ortho_hits:
                        try:
                            neighbor_dict[core_tax][mirid] = (
                                ('inside', ortho_hits[core_tax])
                            )
                        except:
                            neighbor_dict[core_tax] = (
                                {mirid: ('inside', ortho_hits[core_tax])}
                            )
                    break
                # case 4.2): miRNA opposite of gene
                elif (
                        start >= gene_data[1]
                        and end <= gene_data[2]
                        and strand != gene_data[3]
                ):
                    solved = True
                    print(
                        '{0} is located opposite of the gene {1}.'
                            .format(mirid, gene_data[0])
                    )
                    ortho_hits = ortho_search(gene_data[0], ortho_dict)
                    for core_tax in ortho_hits:
                        try:
                            neighbor_dict[core_tax][mirid] = (
                                ('opposite', ortho_hits[core_tax])
                            )
                        except:
                            neighbor_dict[core_tax] = (
                                {mirid: ('opposite', ortho_hits[core_tax])}
                            )
                    break
                # case 4.3): miRNA between genes
                elif (
                        int(ref_dict[chromo][gene][2]) < start
                        and ref_dict[chromo][gene + 1][1] > end
                ):
                    solved = True
                    ###############################################################################
                    print(
                        '{1} is the left neighbor of {2}.'
                        .format(gene, ref_dict[chromo][gene][0], mirid)
                    )
                    print(
                        '{1} is the right neighbor of {2}.'
                        .format(gene, ref_dict[chromo][gene + 1][0], mirid)
                    )
                    left_hits = ortho_search(gene_data[0], ortho_dict)
                    right_hits = (
                        ortho_search(ref_dict[chromo][gene + 1][0], ortho_dict)
                    )
                    # save only the hits where both genes have orthologs in a species
                    if left_hits:
                        # print(left_hits)
                        # print(right_hits)
                        for taxon in left_hits:
                            if taxon in right_hits:
                                try:
                                    neighbor_dict[taxon][mirid] = (
                                        (
                                            'in-between',
                                            [left_hits[taxon],
                                             right_hits[taxon]]
                                        )
                                    )
                                except:
                                    neighbor_dict[taxon] = (
                                        {mirid: (
                                            'in-between',
                                            [left_hits[taxon],
                                             right_hits[taxon]]
                                        )}
                                    )
                            else:
                                print('Orthologs were not found for both flanking genes')
                    break
            if not solved:
                print('Unable to resolve synteny for {}.'.format(mirid))

    #################################################################################################

    # Search for the coordinates of the orthologs and extract the sequences
    for taxon in neighbor_dict:
        print('\n### Starting synteny analysis for {}'.format(taxon))
        print('# Trying to parse annotation file for {}.'.format(taxon))
        fe = path_dict[taxon]['annotation'].split('.')[-1]
        if fe == 'gtf':
            core_dict = gtf_parser(path_dict[taxon]['annotation'])
        elif fe in ['gff3', 'gff']:
            core_dict = gff_parser(path_dict[taxon]['annotation'], id_t)
        else:
            print('ERROR: File type "{}" not valid as reference annotation'.format(ft))
            sys.exit()

        print('# Loading genome file')
        fasta_path = path_dict[taxon]['genome']
        core_gen_dict = '{}/core_genomes'.format(output)
        if not os.path.isdir(core_gen_dict):
            os.mkdir(core_gen_dict)
        slink = '{}/slink_to_{}'.format(core_gen_dict, taxon)
        try:
            os.symlink(fasta_path, slink)
        except FileExistsError:
            pass
        genome = pyfaidx.Fasta(slink)
        print('Done')

        # print('YOU MADE IT THIS FAR.')
        for mirna in neighbor_dict[taxon]:
            # print(mirna)
            style = neighbor_dict[taxon][mirna][0]
            if style == 'inside' or style == 'opposite':
                print(f'# {mirna} location: {style} of gene')
                ###############################################################################
                try:
                    ortho_data = (
                        core_dict[neighbor_dict[taxon][mirna][1]]
                    )
                except KeyError as e:
                    print('{} not found in annotation file.'.format(e))
                    continue
                positions = list(
                    core_dict[ortho_data[0]][ortho_data[1]][1:4]
                )
                coordinates = [ortho_data[0]] + positions
                seq = (
                    genome[coordinates[0]]
                    [coordinates[1] - 1:coordinates[2]].seq
                )
                # print(seq[0:10])
                try:
                    mirna_dict[mirna][taxon] = seq
                except KeyError:
                    mirna_dict[mirna] = {taxon: seq}
            elif style == 'in-between':
                print(f'# {mirna} location: between genes')
                try:
                    left_data = (
                        core_dict[neighbor_dict[taxon][mirna][1][0]]
                    )
                    right_data = (
                        core_dict[neighbor_dict[taxon][mirna][1][1]]
                    )
                except KeyError as e:
                    print('{} not found in annotation file.'.format(e))
                    continue
                # Test to see if the two orthologs are themselves neighbors where their
                # distance cannot be larger than the selected mgi value. This accounts
                # for insertions in the core species.
                # TODO: Apply mgi also to the reference species to account for insertions
                # in the reference.
                #                 print(f'left_data: {left_data}')
                #                 print(f'right_data: {right_data}')
                if (
                        left_data[0] == right_data[0]
                        and abs(left_data[1] - right_data[1]) <= mgi
                ):
                    # Determine which sequence to include for the synteny-based ortholog search
                    # depending on the order of orthologs. The order of the orthologs in the core
                    # species might be inverted compared to that in the reference species.
                    ###############################################################################
                    if left_data[1] < right_data[1]:
                        # print('left')
                        # print(core_gtf_dict[left_data[0]][left_data[1]])
                        # print(core_gtf_dict[right_data[0]][right_data[1]])
                        contig = left_data[0]
                        # print(contig)
                        seq_start = int(
                            core_dict[left_data[0]][left_data[1]][2]
                        )
                        # print(seq_start)
                        seq_end = (
                            core_dict[right_data[0]][right_data[1]][1]
                        )
                        # print(seq_end)
                        seq = genome[contig][seq_start - 1:seq_end].seq
                        try:
                            mirna_dict[mirna][taxon] = seq
                        except:
                            mirna_dict[mirna] = {taxon: seq}
                    elif right_data[1] < left_data[1]:
                        # print('right')
                        # print(core_gtf_dict[left_data[0]][left_data[1]])
                        # print(core_gtf_dict[right_data[0]][right_data[1]])
                        contig = left_data[0]
                        # print(contig)
                        # seq_start = int(
                        #     core_gtf_dict[left_data[0]][left_data[1]][2]
                        # )
                        # print(seq_start)
                        # seq_end = (
                        #     core_gtf_dict[right_data[0]][right_data[1]][1]
                        # )
                        seq_start = int(
                            core_dict[right_data[0]][right_data[1]][2]
                        )
                        # print(seq_start)
                        seq_end = (
                            core_dict[left_data[0]][left_data[1]][1]
                        )
                        # print(seq_end)
                        seq = genome[contig][seq_start - 1:seq_end].seq
                        try:
                            mirna_dict[mirna][taxon] = seq
                        except:
                            mirna_dict[mirna] = {taxon: seq}
                    print('Synteny fulfilled.')
                else:
                    print(
                        'No shared synteny for {} in {}.'
                            .format(mirna, taxon)
                    )
                    continue
                    # print(left_data)
                    # print(right_data)
            else:
                print('## Neither inside, opposite, nor in-between')
                # print(neighbor_dict[taxon][mirna])
                continue
            print('Candidate region found')

    # Write output file
    for mirna in mirna_dict:
        # print('{0}/{1}/{1}.fa'.format(output, mirna))
        with open('{0}/{1}/{1}.fa'.format(output, mirna), 'w') as outfile:
            for core_taxon in mirna_dict[mirna]:
                outfile.write(
                    '>{0}\n{1}\n'
                    .format(core_taxon, mirna_dict[mirna][core_taxon])
                )
    if not mirna_dict:
        print('\nWARNING: No syntenic regions found in the core species')

    for mirna in mirnas:
        print('\n### Starting reciprocal BLAST process')
        sto_path = blastsearch(mirna, ref_genome, output, cpu, dust)

        if create_model == 'yes' and sto_path is not None:
            print('### Starting to construct covariance model from alignment')
            model_dir = f'{output}/CMs'
            if not os.path.isdir(model_dir):
                os.mkdir(model_dir)
            model_out = f'{model_dir}/{mirna[0]}.cm'
            if not os.path.isfile(model_out):
                # Initiate covariance model construction and calibration.
                cmc = CmConstructor(sto_path, model_dir, mirna[0], cpu)
                # Construct the model.
                cmc.construct()
                # Calibrate the model.
                cmc.calibrate()
            else:
                print(f'Model of {mirna[0]} already found at {model_dir}. Nothing done..')

    print('\n### Construction of core set finished')


if __name__ == '__main__':
    main()



"""
ncOrtho - Targeted ortholog search for miRNAs
Copyright (C) 2021 Felix Langschied

ncOrtho is a free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

ncOrtho is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ncOrtho.  If not, see <http://www.gnu.org/licenses/>.
"""

import pyfaidx
import subprocess as sp
import os
import sys


def candidate_blast(db, c, evalue, seq, task):
    # extract candidate regions
    print('# Identifying candidate regions for cmsearch heuristic')
    blast_command = (
        'blastn -task {0} -db {1} '
        '-num_threads {2} -evalue {3} '
        '-outfmt "6 sseqid sstart send sstrand length"'.format(task, db, c, evalue)
    )
    blast_call = sp.Popen(
        blast_command, shell=True, stdin=sp.PIPE, stdout=sp.PIPE, stderr=sp.PIPE, encoding='utf8'
    )
    res, err = blast_call.communicate(seq)
    return res, err



def cmsearcher(mirna, cm_cutoff, cpu, msl, models, query, blastdb, out, cleanup, heuristic):
    """
    Parameters
    ----------
    mirna       :   Mirna object
    cm_cutoff   :   Minimum bit score of the CMsearch hit relative to the
                    bit score that results from searching in the reference genome with the same model
    cpu         :   Number of cores to use
    msl         :   Minimum length of CMsearch hit relative to the reference pre-miRNA length
    models      :   Path to the directory that contains the CMs
    query       :   Path to the query genome
    blastdb     :   Optional Path to a BLASTdb of the query genome
    out         :   Path to the output directory
    cleanup     :   Delete intermediate files True/False
    heuristic   :   Should heuristic mode be used and if yes, set parameters (True/False, evalue, minlength)

    Returns
    -------
    cm_results : Dictionary containing hits of the cmsearch

    """
    mirna_id = mirna.name
    cm_results = False
    # Calculate the bit score cutoff.
    cut_off = mirna.bit * cm_cutoff
    # Calculate the length cutoff.
    len_cut = len(mirna.pre) * msl
    blast_len_cut = len(mirna.pre) * heuristic[2]
    # if no model exists for mirna return False
    if not os.path.isfile('{}/{}.cm'.format(models, mirna_id)):
        print('# No model found for {}. Skipping..'.format(mirna_id))
        exitstatus = 'No covariance model found'
        return cm_results, exitstatus
    # Perform covariance model search.
    # Report and inclusion thresholds set according to cutoff.

    if heuristic[0]:
        # extract candidate regions
        #candidate_blast(db, c, evalue, seq, task):
        res, err = candidate_blast(blastdb, cpu, heuristic[1], mirna.pre, 'blastn')
        if not err and not res:
            print('No BLASTn candidate regions with pre-miRNA. Trying mature sequence')
            res, err = candidate_blast(blastdb, cpu, 10, mirna.mature, 'blastn-short')
            # turn off length blast length cutoff
            blast_len_cut = 0
        if err:
            print(f'ERROR: {err}')
            exitstatus = 'ERROR during candidate region BLAST search'
            return cm_results, exitstatus

        # print('Blast step finished')
        result = list(filter(None, res.split('\n')))
        hit_list = [line.split() for line in result if line and float(line.split()[-1]) >= blast_len_cut]
        if not hit_list:
            exitstatus = 'No BLASTn candidate regions above length cutoff'
            return cm_results, exitstatus
        # hit_list = [line.split() for line in result if line]
        print(f'# Found {len(hit_list)} BLAST hits of the reference '
              f'pre-miRNA in the query')
        genes = pyfaidx.Fasta(query)
        # collect heuristic sequences
        # print('collecting sequences')
        # set border regions to include in analysis
        extraregion = 1000
        diff_dict = {}
        hit_at_start = False
        hit_at_end = False
        heuristic_fa = '{0}/candidate_region_{1}.out'.format(out, mirna_id)
        with open(heuristic_fa, 'w') as of:
            for hit in hit_list:
                chrom, start, end, strand, length = hit
                strand = strand.replace('plus', '+')
                strand = strand.replace('minus', '-')
                if strand == '-':
                    start, end = end, start
                header = ">{}|{}|{}|{}\n".format(chrom, start, end, strand)
                # print(header)
                start = int(start)
                end = int(end)
                if start > extraregion:
                    n_start = start - extraregion
                else:
                    # hit starts at the beginning of the chromosome
                    n_start = 0
                    hit_at_start = True
                n_end = end + extraregion
                if n_end > genes[chrom][-1].end:
                    # hit is at the end of the chromosome
                    n_end = genes[chrom][-1].end
                    hit_at_end = True
                if strand == '+':
                    sequence = genes[chrom][n_start:n_end].seq
                elif strand == '-':
                    sequence = genes[chrom][n_start:n_end].reverse.complement.seq
                header = ">{}|{}|{}|{}|{}|{}\n".format(chrom, start, end, strand, hit_at_start, hit_at_end)
                of.write(header)
                of.write(f'{sequence}\n')

        cm_results = []
        return_data = []
        cms_command = (
            'cmsearch --cpu {0} --noali -T {3} --incT {3} {1}/{2}.cm {4}'
            .format(cpu, models, mirna_id, cut_off, heuristic_fa)
        )
        cms_call = sp.Popen(
            cms_command, shell=True, stdout=sp.PIPE, stderr=sp.PIPE, encoding='utf8'
        )
        res, err = cms_call.communicate(mirna.pre)
        # delete temporary file
        if cleanup:
            os.remove(heuristic_fa)
        # read results
        if err:
            print(f'ERROR: {err}')
            sys.exit()
        for line in res.split('\n'):
            if line.startswith('  ('):
                if 'No hits detected that satisfy reporting thresholds' in line:
                    exitstatus = 'CMsearch found not hits'
                    return cm_results, exitstatus
                else:
                    data = line.strip().split()
                    cm_results.append(data)
        if not cm_results:
            exitstatus = 'CMsearch found no hits'
            return cm_results, exitstatus

        for data in cm_results:
            # print(data)
            # parse CMsearch output of candidate regions to acutal genomic regions
            blast_chrom = data[5].split('|')[0]
            blast_start, blast_end = map(int, data[5].split('|')[1:3])
            hit_at_start, hit_at_end = data[5].split('|')[4:6]
            blast_strand = data[5].split('|')[3]
            cm_start, cm_end = map(int, data[6:8])
            # cm_start, cm_end = map(int, data[5:7])
            cm_strand = data[8]
            if cm_strand == '-':
                # reverse hits should not be possible since blasthits were extracted
                # as reverse complement if they were on the minus strand
                print('Unexpected hit of CMsearch')
                continue
            if hit_at_start == 'True':
                print('# cmsearch hit for {} at the beginning of a chromosome in the query species'.format(mirna_id))
                hit_start = cm_start
                hit_end = cm_end
            elif hit_at_end == 'True':
                print('# cmsearch hit for {} at the end of a chromosome in the query species'.format(
                    mirna_id))
                hit_start = blast_start + (cm_start - extraregion) - 1
                hit_end = hit_start + (cm_end - cm_start) - 1
            else:
                hit_start = blast_start + (cm_start - extraregion) - 1
                hit_end = hit_start + (cm_end - extraregion) - 1
            # fill output variable
            return_data.append([blast_chrom, hit_start, hit_end, blast_strand, float(data[3])])
        if not return_data:
            exitstatus = 'No hits left after parsing of CMsearch results'
            return False, exitstatus

        cm_results, exitstatus = cmsearch_parser(return_data, cut_off, len_cut, mirna_id)

        if not cleanup:
            heur_results = '{0}/cmsearch_{1}.out'.format(out, mirna_id)
            with open(heur_results, 'w') as of:
                of.write('# CMsearch hits in query genome\n')
                for hit in cm_results:
                    out_d = [str(entry) for entry in cm_results[hit]]
                    of.write('\t'.join(out_d))
                    of.write('\n')

    # non heuristic mode (searches whole query genome with CM)
    else:
        cm_res_list = []
        cms_output = '{0}/cmsearch_{1}.out'.format(out, mirna_id)
        if not os.path.isfile(cms_output):
            # heuristic_cms = '{0}/cmsearch_heuristic_{1}.out'.format(out, mirna_id)
            print('# Running covariance model search for {}'.format(mirna_id))
            cms_command = (
                'cmsearch -T {5} --incT {5} --cpu {0} --noali '
                '-o {1} {2}/{3}.cm {4}'
                .format(cpu, cms_output, models, mirna_id, query, cut_off)
            )
            sp.run(cms_command, shell=True)

        else:
            print('# Found cm_search results at: {}. Using those'.format(cms_output))
        data = []
        with open(cms_output, 'r') as inf:
            for line in inf:
                if line.startswith('  ('):
                    if 'No hits detected that satisfy reporting thresholds' in line:
                        exitstatus = 'CMsearch found not hits'
                        return cm_results, exitstatus
                    data = line.strip().split()
                    h_chrom = data[5]
                    h_start = int(data[6])
                    h_end = int(data[7])
                    h_score = float(data[3])
                    h_strand = data[8]
                    cm_res_list.append([h_chrom, h_start, h_end, h_strand, h_score])
        if not data:
            exitstatus = 'cmSearch did not find anything'
            return cm_results, exitstatus

        cm_results, exitstatus = cmsearch_parser(cm_res_list, cut_off, len_cut, mirna_id)
    return cm_results, exitstatus


# cmsearch_parser: Parse the output of cmsearch while eliminating
#                  duplicates and filtering entries according to the
#                  defined cutoff.
def cmsearch_parser(cms, cmc, lc, mirid):
    """
    Parse the output of cmsearch while eliminating duplicates and filtering
    entries according to the defined cutoff.

    Parameters
    ----------
    cms     :   List with CMsearch hits [chrom, start, end, strand, score]
    cmc     :   Cutoff to decide which candidate hits should be included for the reverse BLAST search
    lc      :   Length cutoff
    mirid   :   miRNA ID

    Returns
    -------
    hits_dict : Dictionary containing hits of the cmsearch

    """
    # Output
    hits_dict = {}
    # Required for finding duplicates, stores hits per chromosome
    chromo_dict = {}
    cut_off = cmc

    for candidate_nr, hit in enumerate(cms, 1):
        h_chrom, h_start, h_end, h_strand, h_score = hit
        # blastparser expects the start to be the smaller number, will extract reverse complement if on - strand
        if h_start > h_end:
            start_tmp = h_start
            h_start = h_end
            h_end = start_tmp
        data = (
            '{0}_c{1}'.format(mirid, candidate_nr),
            h_chrom, h_start, h_end, h_strand, h_score
        )
        hits_dict[data[0]] = data

        # Store the hits that satisfy the bit score cutoff to filter
        # duplicates.
        try:
            chromo_dict[data[1]].append(data)
        except:
            chromo_dict[data[1]] = [data]

    # Loop over the candidate hits to eliminate duplicates.
    for chromo in chromo_dict:
        nrhits = len(chromo_dict[chromo])
        if nrhits > 1:
            for hitnr in range(nrhits):
                start = int(chromo_dict[chromo][hitnr][2])
                stop = int(chromo_dict[chromo][hitnr][3])
                strand = chromo_dict[chromo][hitnr][4]
                score = float(chromo_dict[chromo][hitnr][5])
                for chitnr in range(hitnr +1, nrhits):
                    if strand != chromo_dict[chromo][chitnr][4]:
                        cstart = int(chromo_dict[chromo][chitnr][2])
                        cstop = int(chromo_dict[chromo][chitnr][3])
                        cscore = float(chromo_dict[chromo][chitnr][5])
                        # Test if the two hits from opposite strands
                        # overlap, which means one of them is
                        # (probably) a false positive.
                        # Out of two conflicting hits, the one with the
                        # highest cmsearch bit score is retained.
                        if (
                                start in range(cstart, cstop +1)
                                or stop in range(cstart, cstop +1)
                                or cstart in range(start, stop +1)
                                or cstop in range(start, stop +1)
                        ):
                            if score > cscore:
                                try:
                                    del hits_dict[chromo_dict[chromo]
                                    [chitnr][0]]
                                except:
                                    pass
                            else:
                                try:
                                    del hits_dict[chromo_dict[chromo]
                                    [hitnr][0]]
                                except:
                                    pass
    exitstatus = 'Sucess'
    return hits_dict, exitstatus



import subprocess as sp
import os
import glob
import sys

from ncOrtho.utils import check_blastdb
from ncOrtho.utils import make_blastndb


def make_alignment(out, mirna, cpu, core):
    alignment = '{}/{}.aln'.format(out, mirna)
    stockholm = '{}/{}.sto'.format(out, mirna)

    # Call T-Coffee for the sequence alignment.
    print('Building T-Coffee alignment.')
    tc_cmd_1 = (
        't_coffee -quiet -multi_core={} -special_mode=rcoffee -in {} '
        '-output=clustalw_aln -outfile={}'
            .format(cpu, core, alignment)
    )
    sp.call(tc_cmd_1, shell=True)

    # Extend the sequence-based alignment by structural information.
    # Create Stockholm alignment.
    print('Adding secondary structure to Stockholm format.')
    tc_cmd_2 = (
        't_coffee -other_pg seq_reformat -in {} -action +add_alifold -output '
        'stockholm_aln -out {}'
            .format(alignment, stockholm)
    )
    sp.call(tc_cmd_2, shell=True)
    return stockholm

# Perform reciprocal BLAST search and construct Stockholm alignment
def blastsearch(mirna, r_path, o_path, c, dust):
    """

    Parameters
    ----------
    m_path  :   Path to file with miRNA information
    r_path  :   Path to reference genome.
    o_path  :   Output Name
    c       :   Number of Threads for BLAST search.
    dust    :   Use dust filter for BLAST searches (yes/no)

    Returns
    -------
    Path to Alignment of core pre-miRNAs in Stockholm format

    """
    core_set = {}


    # with open(m_path) as infile:
    #     mirnas = [
    #         line.strip().split()
    #         for line in infile
    #         if not line.startswith('#')
    #     ]
    # for mirna in mirnas:
    mirid = mirna[0]
    # this fasta file is created by the the main() script
    fasta = '{0}/{1}/{1}.fa'.format(o_path, mirid)

    # Ensure that output folder exists and change to this folder.
    out_folder = '{}/{}'.format(o_path, mirid)
    if not os.path.isdir(out_folder):
        os.mkdir(out_folder)

    os.chdir('{}/{}'.format(o_path, mirid))
    print('\n### {} ###'.format(mirid))
    # Coordinates of the ncRNA
    mchr = mirna[1].replace('chr', '')
    mstart = int(mirna[2])
    mend = int(mirna[3])

    # Start of reference bit score computation.
    # Convert RNA sequence into DNA sequence.
    preseq = mirna[5].replace('U', 'T')

    if not os.path.isfile(fasta):
        # no synteny region detected, make alignment of reference miRNA only
        with open(fasta, 'w') as fastah:
            fastah.write(f'>{mirid}\n{preseq}\n')
        stock = make_alignment(out_folder, mirid, c, fasta)
        return stock

    # The miRNA precursors can show a low level of complexity, hence it is
    # required to deactivate the dust filter for the BLAST search.


    # check if blastdb of reference genome exists
    fname = '.'.join(r_path.split("/")[-1].split('.')[0:-1])
    ref_blastdb = f'{o_path}/refBLASTdb/{fname}'
    if not check_blastdb(ref_blastdb):
        make_blastndb(r_path, ref_blastdb)

    bit_check = (
        'blastn -num_threads {0} -dust {1} -task megablast -db {2} '
        '-outfmt \"6 bitscore\"'.format(c, dust, ref_blastdb)
    )
    print('# Determining reference bit score..')
    ref_bit_cmd = sp.Popen(
        bit_check, shell=True, stdin=sp.PIPE,
        stdout=sp.PIPE, stderr=sp.STDOUT, encoding='utf8'
    )
    ref_results, err = ref_bit_cmd.communicate(preseq)
    if err:
        print(err)
    if not ref_results:
        print('WARNING: Reference sequence of {} not found in reference Genome. '
              'Consider turning the dust filter off'.format(mirid))
        skip_file = '{}/not_found_in_ref.fa'.format(o_path)
        if not os.path.isfile(skip_file):
            with open(skip_file, 'w') as of:
                of.write('>{}\n{}\n'.format(mirid, preseq))
        else:
            with open(skip_file, 'a') as of:
                of.write('>{}\n{}\n'.format(mirid, preseq))
        return None
    ref_bit_score = float(ref_results.split('\n')[0].split('\t')[0])
    print(f'BLAST bit score of reference miRNA against reference genome: {ref_bit_score}')

    print('# Trying to find pre-miRNA candidates in syntenic region')
    # Check if BLAST database already exists, otherwise create it.
    file_extensions = ['.nhr', '.nin', '.nsq']
    for fe in file_extensions:
        checkpath = '{}{}'.format(fasta, fe)
        if not os.path.isfile(checkpath):
            print(
                'Constructing BLAST database of candidate regions.'
            )
    # At least one of the BLAST db files is not existent and has to be
    # created.
            db_command = (
                'makeblastdb -dbtype nucl -in {}'
                .format(fasta)
            )
            sp.call(db_command, shell=True)
            break

    blastn_cmd = (
        'blastn -num_threads {0} -task blastn -dust {1} -db {2} -outfmt \"6 '
        'sseqid evalue bitscore sseq\"'.format(c, dust, fasta)
    )
    blastn = sp.Popen(
        blastn_cmd, shell=True, stdin=sp.PIPE,
        stdout=sp.PIPE, stderr=sp.PIPE, encoding='utf8'
    )
    results, err = blastn.communicate(preseq)
    if err:
        print(err)
        sys.exit()


    ##### Collect best hit for each core set species if it is within the accepted bit score range
    core_dict = {}
    result_list = results.split('\n')
    # print(result_list)
    if result_list:
        #print(result_list)
        for hit in result_list:
            if hit:
                hit_data = hit.split()
                # print(hit_data)
                # print(f'hit bitscore: {hit_data[2]}')
                # print(f'you needed: {0.5*ref_bit_score}')
                # print(f'and you got: {hit_data[2]}')
                if (
                    not hit_data[0] in core_dict
                    and float(hit_data[2]) >= 0.5*ref_bit_score
                ):
                    core_dict[hit_data[0]] = hit
                    print(f'pre-miRNA candidate found for {hit_data[0]}! ({hit_data[2]}/{0.5*ref_bit_score})')
                    break
                else:
                    print(f'Syntenic candidate region BLAST hit below reference bit score threshold ({hit_data[2]}/{0.5*ref_bit_score})')

    if len(core_dict) == 0:
        print('WARNING: No region in the core species scored above the reference bit score threshold')


    ##### Re-BLAST #####
    print('\n### Starting reciprocal BLAST search.')
    accept_dict = {}
    for species in core_dict.keys():
        print(f'# {species}')
        # Make sure to eliminate gaps
        candidate_seq = core_dict[species].split()[3].replace('-', '')
        reblastn_cmd = (
            'blastn -num_threads {0} -task blastn -dust {1} -db {2} -outfmt \"6'
            ' sseqid sstart send evalue bitscore\"'
            .format(c, dust, ref_blastdb)
        )
        reblastn = sp.Popen(
            reblastn_cmd, shell=True, stdin=sp.PIPE,
            stdout=sp.PIPE, stderr=sp.STDOUT, encoding='utf8'
        )
        reresults, reerr = reblastn.communicate(candidate_seq)
        if reerr:
            print(reerr)
            sys.exit()

    # Check if reverse hit overlaps with reference miRNA
        if reresults:
            first_hit = reresults.split('\n')[0].split()
            rchr = first_hit[0]
            rstart = int(first_hit[1])
            rend = int(first_hit[2])
            if rchr == mchr:
                # print('Same chromosome.')
                if (
                    (rstart <= mstart and mstart <= rend)
                    or (rstart <= mend and mend <= rend)
                ):
                    #first within second
                    print('Reciprocity fulfilled.')
                    accept_dict[species] = core_dict[species]
                elif (
                    (mstart <= rstart and rstart <= mend)
                    or (mstart <= rend and rend <= mend)
                ):

                    # second within first
                    print('Reciprocity fulfilled.')
                    accept_dict[species] = core_dict[species]
                else:
                    print('Reciprocity unfulfilled.')
            else:
                print('Hit on chromosome {}, expected {}'.format(rchr, mchr))
        else:
            print(
                'No reverse hit for {}. Reciprocity unfulfilled.'
                .format(mirid)
            )
    # write intermediate output file
    corefile = '{}/{}_core.fa'.format(out_folder, mirid)
    with open(corefile, 'w') as outfile:
        outfile.write('>{}\n{}\n'.format(mirid, preseq))
        print('\n### Starting Alignment')
        # print('>{}\n{}'.format(mirid, preseq))
        for accepted in accept_dict:
            # print('>{}\n{}'.format(accepted, core_dict[accepted].split('\t')[3]))
            outfile.write(
                '>{}\n{}\n'
                    .format(accepted, accept_dict[accepted].split('\t')[3])
                    .replace('-', '')
            )
    stock = make_alignment(out_folder, mirid, c, corefile)
    return stock





"""
ncOrtho - Targeted ortholog search for miRNAs
Copyright (C) 2021 Felix Langschied

ncOrtho is a free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

ncOrtho is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ncOrtho.  If not, see <http://www.gnu.org/licenses/>.
"""

import subprocess as sp
from Bio import AlignIO
from Bio.Phylo.TreeConstruction import DistanceCalculator
import os
import sys


def calculate_distance_matrix(aln_path):
    # align reBLAST hits
    aln_cmd = (
        't_coffee {0} -no_warning -quiet -type=dna '
        ' -output=fasta_aln -outfile={0}'.format(aln_path)
    )
    sp.run(aln_cmd, shell=True)
    aln = AlignIO.read(open(aln_path), 'fasta')
    calculator = DistanceCalculator('identity')
    dm = calculator.get_distance(aln)
    # delete temporary files
    os.remove(aln_path)
    curr_path = os.getcwd()
    os.remove('{}/{}.dnd'.format(curr_path, aln_path.split('/')[-1].split('.fa')[0]))
    return dm


# BlastParser object performs reverse BLAST search and reports if a candidate
# fulfills the reverse best hit criterion
class BlastParser(object):
    # init parameters:
    # mirna: Mirna object that holds the location for the reference miRNA
    # blastpath: path to the output file of the reverse BLAST search
    # msl: ncOrtho minimum sequence length threshold
    def __init__(self, mirna, blasthits, msl):
        self.start = mirna.start
        self.end = mirna.end
        self.chromosome = mirna.chromosome
        self.strand = mirna.strand
        self.refseq = mirna.pre
        del mirna
        self.blasthits = blasthits
        self.msl = msl

    def evaluate_besthit(self,):
        # with open(self.blastpath) as blastfile:
        #     blasthits = [line.strip().split() for line in blastfile]
        # If no BLAST hit was found, the search failed by default.
        if not self.blasthits:
            print('Rejecting: No reciprocal BLAST hit found')
            return False
        # Otherwise, check if the best hit and the reference miRNA overlap.
        else:
            # Gather coordinates of best BLAST hit.
            tophit = self.blasthits[0]
            sseqid = tophit[0]
            sstrand = tophit[3]
            if sstrand == 'plus':
                sstart = int(tophit[1])
                send = int(tophit[2])
            elif sstrand == 'minus':
                sstart = int(tophit[2])
                send = int(tophit[1])
            else:
                print('ERROR: re-BLAST on neither plus nor minus strand')
                print(tophit)
                sys.exit()

            # del blasthits
            # Sequences must be on the same contig, otherwise overlap can be
            # ruled out instantaneously
            if not sseqid == self.chromosome:
                print(f'Rejecting: Contig/Chromosome does not match. Expected {self.chromosome} but found {sseqid}')
                return False
            # Contigs match, so overlap is possible.
            else:
                print(f'Start miRNA: {self.start} End miRNA: {self.end}')
                print(f'Start BLAST hit: {sstart} End BLAST hit: {send}')
                # first within second
                if (
                    (sstart <= self.start <= send)
                    or (sstart <= self.end <= send)
                ):
                    return True
                # second within first
                elif (
                    (self.start <= sstart <= self.end)
                    or (self.start <= send <= self.end)
                ):
                    return True
                # No overlap
                else:
                    print('Rejecting: No overlap between miRNA and best BLAST hit')
                    return False

    def check_coortholog_ref(self, candidate_seq, out):
        # 1) get query sequence
        # 2) get reference sequence
        # 3) get sequence of best blast hit
        # 4) align sequences
        # 5) calculate difference
        # 6) if distance between best blast hit and reference smaller than
        # difference between best hit and candidate, accept candidate -> reBLAST hit co-ortholog instead of reference

        # get process id to ensure that the correct library is used when running ncOrtho in parallel
        pid = os.getpid()
        tmp_out = '{0}/{1}.fa'.format(out, pid)
        # extract sequences of reBLAST hits
        with open(tmp_out, 'w') as of:
            of.write('>candidate\n')
            of.write('{}\n'.format(candidate_seq))
            of.write('>reference\n')
            of.write('{}\n'.format(self.refseq))
            of.write('>best_hit\n')
            of.write(self.blasthits[0][12].replace('-', ''))

        dm = calculate_distance_matrix(tmp_out)
        distance_hit_query = dm['best_hit', 'candidate']
        distance_ref_hit = dm['best_hit', 'reference']

        if distance_ref_hit < distance_hit_query:
            print(
                "\t Distance query - BLAST hit: %6.4f, Distance blast hit - reference: %6.4f\tAccepting\n"
                %(distance_hit_query, distance_ref_hit)
            )
            return True
        else:
            print(
                "\t Distance query - BLAST hit: %6.4f, Distance blast hit - reference: %6.4f\tRejecting\n"
                %(distance_hit_query, distance_ref_hit)
            )
            return False


class ReBlastParser(object):
    def __init__(self, mirna, reblast_dict):
        self.refseq = mirna.pre
        self.hits = reblast_dict
        self.best_candidate = list(reblast_dict.keys())[0]

    def verify_coorthologs(self, out):
        out_dict = {}
        # get process id to ensure that the correct library is used when running ncOrtho in parallel
        pid = os.getpid()
        tmp_out = '{0}/{1}.fa'.format(out, pid)
        with open(tmp_out, 'w') as of:
            of.write('>reference\n')
            of.write(f'{self.refseq}\n')
            for candidate in self.hits:
                of.write(f'>{candidate}\n')
                of.write(f'{self.hits[candidate]}\n')

        dm = calculate_distance_matrix(tmp_out)
        for candidate in self.hits:
            if candidate == self.best_candidate:
                out_dict[candidate] = self.hits[candidate]
            else:
                distance_candidates = dm[self.best_candidate, candidate]
                distance_best_ref = dm[self.best_candidate, 'reference']
                if distance_candidates < distance_best_ref:
                    print(
                        f'co-ortholog detected: distance of best candidate to {candidate} {distance_candidates} '
                        f'compared to distance of best candidate to reference of {distance_best_ref}'
                    )
                    out_dict[candidate] = self.hits[candidate]
        return out_dict



import glob
import os
import tempfile
import subprocess as sp
import sys
import shutil
import inspect
from collections import Counter
import argparse

"""
Species in spec_to_skip will not be part of multiple sequence alignments
"""


def make_supermatrix(out):
    align_out = '{}/alignments'.format(out)
    tree_out = '{}/supermatrix'.format(out)
    # curr_dir = sys.path[0]
    curr_dir = '/'.join(inspect.getfile(inspect.currentframe()).split('/')[:-1])
    if not os.path.isdir(tree_out):
        os.mkdir(tree_out)
    print('# Creating supermatrix alignment')
    cmd = (
        'perl {}/concat_alignments_dmp.pl -in {} -out supermatrix.aln'
            .format(curr_dir, align_out)
    )
    res = sp.run(cmd, shell=True, capture_output=True)
    if res.returncode != 0:
        print('ERROR:')
        print(res.stderr.decode('utf-8'))
        sys.exit()
    # print('# Done')

    print('# De-gapping alignment')
    cmd = (
        'perl {}/degapper.pl -in {}/supermatrix.aln'.format(curr_dir, out)
    )
    sp.run(cmd, shell=True)
    # move files to output dir
    shutil.move(f'{out}/supermatrix.aln', f'{tree_out}/supermatrix.aln')
    shutil.move(f'{out}/subalignment_positions.txt', f'{tree_out}/subalignment_positions.txt')
    shutil.move(f'{out}/supermatrix.aln.proc', f'{tree_out}/supermatrix.aln.proc')
    # print('# Done')
    # return f'{tree_out}/supermatrix.aln'
    return f'{tree_out}/supermatrix.aln.proc'


def main():
    # Print header
    print('\n'+'#'*39)
    print('###'+' '*33+'###')
    print('###   Analysis of ncOrtho results   ###')
    print('###'+' '*33+'###')
    print('#'*39+'\n')

    # Parse command-line arguments
    # Define global variables
    parser = argparse.ArgumentParser(
        description='Analzye ncOrtho results. Gives: PhyloProfile input, supermatrix alignment, species tree.'
    )
    parser._action_groups.pop()
    required = parser.add_argument_group('Required Arguments')
    optional = parser.add_argument_group('Optional Arguments')
    required.add_argument(
        '-r', '--results', metavar='<path>', type=str,
        help='Path to ncOrtho output directory that is to be analyzed'
    )
    required.add_argument(
        '-o', '--output', metavar='<path>', type=str,
        help='Path to location where output of analysis should be stored. '
             'Assumes that result files end with "_orthologs.fa"'
    )
    required.add_argument(
        '-m', '--mapping', metavar='<path>', type=str,
        help='Tab seperated mapping file between NCBI taxonomy id and name of the query species '
             'or more generally the name of the directory that contains the ncOrtho results of each species'
             '(e.g "9606\tHomo_sapiens")'
    )
    ##########################################################################################
    optional.add_argument(
        '--skip', metavar='str', type=str, nargs='?', const='', default='',
        help=(
            'Comma separated list of species for which analyses should be skipped'
        )
    )
    optional.add_argument(
        '--auto_skip', metavar='float', type=float, nargs='?', const=0.0, default=0.0,
        help=(
            'Exclude species for which less than a fraction of the reference '
            'miRNAs were detected from species tree calculation. '
            '(e.g. --auto_skip 0.5 means that only species in which at least 50%% '
            'of the reference miRNAs were found are used for species tree calculation) (Default: Off)'
        )
    )
    optional.add_argument(
        '--iqtree', metavar='str', type=str, nargs='?',
        const='iqtree -s {} -bb 1000 -alrt 1000 -nt AUTO -redo -pre {}/species_tree',
        default='iqtree -s {} -bb 1000 -alrt 1000 -nt AUTO -redo -pre {}/species_tree',
        help=(
            'Call to iqtree. Do not change curly bracket notation '
            '(Default: iqtree -s {} -bb 1000 -alrt 1000 -nt AUTO -redo -pre {}/{})'
        )
    )
    ##########################################################################################
    # Parse arguments
    ##########################################################################################
    if len(sys.argv) == 1:
        parser.print_help()
        sys.exit(1)
    else:
        args = parser.parse_args()

    res_dict = args.results
    taxmap = args.mapping
    outdir = args.output
    spec_to_skip = args.skip.split(',')
    if spec_to_skip[0] == '':
        spec_to_skip.pop()
    iqtree_cmd = args.iqtree
    auto_skip = args.auto_skip

    ##########################################################################################
    # Argument checks
    ##########################################################################################
    if not os.path.isdir(outdir):
        os.mkdir(outdir)
    if auto_skip < 0 or auto_skip > 1:
        print(f'ERROR: --auto_skip must be a value between 0 and 1, not {auto_skip}')
        sys.exit()

    ###############################################################################################
    # Main
    ###############################################################################################
    # find results
    ortholog_files = glob.glob(f'{res_dict}/*/*_orthologs.fa')
    if not ortholog_files:
        ortholog_files = glob.glob(f'{res_dict}/*.fa')
    # read mapping file
    name_2_taxid = {}
    with open(taxmap, 'r') as mp:
        for line in mp:
            taxid, name = line.strip().split('\t')
            name_2_taxid[name] = taxid

    ortho_dict = {}  # ortho_dict = {mirna: {species: seq}}
    # read all results into one dictionary and write phyloprofile input
    pp_in = f'{outdir}/PhyloProfile.long'
    spec_list = []
    with open(pp_in, 'w') as pp:
        pp.write('geneID\tncbiID\torthoID\n')
        for file in ortholog_files:
            with open(file, 'r') as fh:
                for line in fh:
                    if line.startswith('>'):
                        spec, mirna = line.strip().replace('>', '').split('|')[:2]
                    else:
                        seq = line.strip()
                        # write phyloprofile input
                        taxstr = f'ncbi{name_2_taxid[spec]}'
                        group_str = mirna.split('_')[0]
                        pp.write(f'{group_str}\t{taxstr}\t{mirna}\n')
                        # initialize dictionary
                        if group_str not in ortho_dict:
                            ortho_dict[group_str] = {}
                        # only save representative (first) hit
                        if spec not in ortho_dict[group_str]:
                            ortho_dict[group_str][spec] = seq
                            spec_list.append(spec)
    print('# Finished writing PhyloProfile input')

    # count species and skip species with few orthologs before aligning
    if auto_skip:
        spec_counter = Counter(spec_list)
        max_spec_count = spec_counter.most_common(1)[0][1]
        count_cutoff = max_spec_count * auto_skip
        print(f'# Only species with at least {round(count_cutoff)} identified orthologs are used for tree calculation')
        for species, count in dict(spec_counter).items():
            if count < count_cutoff:
                spec_to_skip.append(species)
        print('# Skipping these species:')
        for spsk in spec_to_skip:
            print(spsk)

    # make alignments
    print('# Starting alignments')
    align_out = f'{outdir}/alignments'
    if not os.path.isdir(align_out):
        os.mkdir(align_out)
    for mirna in ortho_dict:
        with tempfile.NamedTemporaryFile(mode='w+') as fp:
            for spec, seq in ortho_dict[mirna].items():
                if spec not in spec_to_skip:
                    fp.write(f'>{spec}\n{seq}\n')
            fp.seek(0)
            aln_cmd = f'muscle -in {fp.name} -out {align_out}/{mirna}.aln'
            res = sp.run(aln_cmd, shell=True, capture_output=True)
            if res.returncode != 0:
                print(res.stderr.decode('utf8'))
                sys.exit()

    superm_path = make_supermatrix(outdir)
    print('# Starting tree calculation')
    tree_out = f'{outdir}/species_tree'
    if not os.path.isdir(tree_out):
        os.mkdir(tree_out)
    tree_cmd = iqtree_cmd.format(superm_path, tree_out)
    res = sp.run(tree_cmd, shell=True, capture_output=True)
    print(res.stdout.decode('utf-8'))
    if res.returncode != 0:
        print(res.stderr.decode('utf-8'))


if __name__ == "__main__":
    main()



"""
ncOrtho - Targeted ortholog search for miRNAs
Copyright (C) 2021 Felix Langschied

ncOrtho is a free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

ncOrtho is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ncOrtho.  If not, see <http://www.gnu.org/licenses/>.
"""


# Modules import
import argparse
import multiprocessing as mp
import os
import subprocess as sp
import sys

# Internal ncOrtho modules
# try:
from ncOrtho.blastparser import BlastParser
from ncOrtho.blastparser import ReBlastParser
from ncOrtho.genparser import GenomeParser
from ncOrtho.cmsearch import cmsearcher
from ncOrtho.utils import check_blastdb
from ncOrtho.utils import make_blastndb
from ncOrtho.utils import find_refbit
# except ImportError:
# from blastparser import BlastParser
# from blastparser import ReBlastParser
# from genparser import GenomeParser
# from cmsearch import cmsearcher
# from utils import check_blastdb
# from utils import make_blastndb
# from utils import find_refbit

###############################################################################


# Central class of microRNA objects
class Mirna(object):
    def __init__(self, name, chromosome, start, end, strand, pre, mature, bit):
        # miRNA identifier
        self.name = name
        # chromosome that the miRNA is located on
        if chromosome.startswith('chr'):
            self.chromosome = chromosome.split('chr')[1]
        else:
            self.chromosome = chromosome
        # start position of the pre-miRNA
        self.start = int(start)
        # end position of the pre-miRNA
        self.end = int(end)
        # sense (+) or anti-sense (-) strand
        self.strand = strand
        # nucleotide sequence of the pre-miRNA
        self.pre = pre.replace('U', 'T')
        # nucleotide sequence of the mature miRNA
        self.mature = mature.replace('U', 'T')
        # reference bit score that miRNA receives by its own
        # covariance model
        self.bit = bit


def mirna_maker(mirpath, cmpath, output, msl):
    """
    Parses the miRNA data input file and returns a dictionary of Mirna objects.

    Parameters
    ----------
    mirpath : STR
        Path to file with microRNA data.
    cmpath : STR
        Path to covariance models.
    output : STR
        Path for writing temporary files.
    msl : FLOAT
        Length filter.

    Returns
    -------
    mmdict : DICT
        Dictionary with Mirna objects.

    """
    
    mmdict = {} # will be the return object
    
    with open(mirpath) as mirna_file:
        mirna_data = [
            line.strip().split() for line in mirna_file
            if not line.startswith('#')
        ]

    print('# Calculating reference bit scores')
    for mirna in mirna_data:
        mirid = mirna[0]
        seq = mirna[5]
        query = '{0}/{1}.fa'.format(output, mirid)
        model = '{0}/{1}.cm'.format(cmpath, mirid)
        # Check if the covariance model even exists, otherwise skip to
        # the next miRNA.
        if not os.path.isfile(model):
            print('WARNING: No covariance model found for {}'.format(mirid))
            mmdict[mirna[0]] = ''
            continue
        # Check if the output folder exists, otherwise create it.
        if not os.path.isdir('{}'.format(output)):
            mkdir = 'mkdir -p {}'.format(output)
            sp.call(mkdir, shell=True)

        # Obtain the reference bit score for each miRNA by applying it
        # to its own covariance model.
        
        # Create a temporary FASTA file with the miRNA sequence as
        # query for external search tool cmsearch to calculate the
        # reference bit score.
        with open(query, 'w') as tmpfile:
            tmpfile.write('>{0}\n{1}'.format(mirid, seq))
        cms_command = (
            'cmsearch -E 0.01 --noali {0} {1}'
            .format(model, query)
        )
        res = sp.run(cms_command, shell=True, capture_output=True)
        if res.returncode == 0:
            top_score = find_refbit(res.stdout.decode('utf-8'))
            mirna.append(top_score)
        else:
            print(f'ERROR: {res.stderr.decode("utf-8")}')
            sys.exit()
        # cleanup
        os.remove(query)
        # Create output.
        mmdict[mirna[0]] = Mirna(*mirna)

    return mmdict


# Allow boolean argument parsing
def str2bool(v):
    if isinstance(v, bool):
        return v
    if v.lower() in ('yes', 'true', 't', 'y', '1'):
        return True
    elif v.lower() in ('no', 'false', 'f', 'n', '0'):
        return False
    else:
        raise argparse.ArgumentTypeError('Boolean value expected.')


# Main function
def main():
    # Print header
    print('\n'+'#'*57)
    print('###'+' '*51+'###')
    print('###   ncOrtho - ortholog search for non-coding RNAs   ###')
    print('###'+' '*51+'###')
    print('#'*57+'\n')

    ##########################################################################################
    # Command line arguments
    ##########################################################################################
    
    # Parse command-line arguments
    # Define global variables
    parser = argparse.ArgumentParser(
        description='Find orthologs of reference miRNAs in the genome of a query species.'
    )
    parser._action_groups.pop()
    required = parser.add_argument_group('Required Arguments')
    optional = parser.add_argument_group('Optional Arguments')
    # covariance models folder
    required.add_argument(
        '-m', '--models', metavar='<path>', type=str, required=True,
        help='Path to directory containing covariance models (.cm)'
    )
    # mirna data
    required.add_argument(
        '-n', '--ncrna', metavar='<path>', type=str, required=True,
        help='Path to Tab separated file with information about the reference miRNAs'
    )
    # output folder
    required.add_argument(
        '-o', '--output', metavar='<path>', type=str, required=True,
        help='Path to the output directory'
    )
    # query genome
    required.add_argument(
        '-q', '--query', metavar='<.fa>', type=str, required=True,
        help='Path to query genome in FASTA format'
    )
    # reference genome
    required.add_argument(
        '-r', '--reference', metavar='<.fa>', type=str, required=True,
        help='Path to reference genome in FASTA format'
    )
    ##########################################################################
    # Optional Arguments
    ##########################################################################
    # query_name
    optional.add_argument(
        '--queryname', metavar='str', type=str, nargs='?', const='', default='',
        help=(
            'Name for the output directory (RECOMMENDED) '
        )
    )
    # cpu, use maximum number of available cpus unless specified otherwise
    optional.add_argument(
        '--cpu', metavar='int', type=int,
        help='Number of CPU cores to use (Default: all available)', nargs='?',
        const=mp.cpu_count(), default=mp.cpu_count()
    )
    # bit score cutoff for cmsearch hits
    optional.add_argument(
        '--cm_cutoff', metavar='float', type=float,
        help='CMsearch bit score cutoff, given as ratio of the CMsearch bitscore '
             'of the CM against the refernce species (Default: 0.5)', nargs='?', const=0.5, default=0.5
    )
    # length filter to prevent short hits
    optional.add_argument(
        '--minlength', metavar='float', type=float,
        help='CMsearch hit in the query species must have at '
             'least the length of this value times the length of the refernce pre-miRNA (Default: 0.7)',
        nargs='?', const=0.7, default=0.7
    )
    optional.add_argument(
        '--heuristic', type=str2bool, metavar='True/False', nargs='?', const=True, default=True,
        help=(
            'Perform a BLAST search of the reference miRNA in the query genome to identify '
            'candidate regions for the CMsearch. Majorly improves speed. (Default: True)'
        )
    )
    optional.add_argument(
        '--heur_blast_evalue', type=float, metavar='float', nargs='?', const=0.5, default=0.5,
        help=(
            'Evalue filter for the BLASTn search that '
            'determines candidate regions for the CMsearch when running ncOrtho in heuristic mode. '
            '(Default: 0.5) (Set to 10 to turn off)'
        )
    )
    optional.add_argument(
        '--heur_blast_length', type=float, metavar='float', nargs='?', const=0.5, default=0.5,
        help=(
            'Length cutoff for BLASTN search with which candidate regions for the CMsearch are identified.'
            'Cutoff is given as ratio of the reference pre-miRNA length '
            '(Default: 0.5) (Set to 0 to turn off)'
        )
    )
    optional.add_argument(
        '--cleanup', type=str2bool, metavar='True/False', nargs='?', const=True, default=True,
        help=(
            'Cleanup temporary files (Default: True)'
        )
    )
    optional.add_argument(
        '--refblast', type=str, metavar='<path>', nargs='?', const='', default='',
        help=(
            'Path to BLASTdb of the reference species'
        )
    )
    optional.add_argument(
        '--queryblast', type=str, metavar='<path>', nargs='?', const='', default='',
        help=(
            'Path to BLASTdb of the query species'
        )
    )
    optional.add_argument(
        '--maxcmhits', metavar='int', nargs='?', const=50, default=50,
        help=(
            'Maximum number of cmsearch hits to examine. '
            'Decreases runtime significantly if reference miRNA in genomic repeat region (Default: 50). '
            'Set to empty variable to disable (i.e. --maxcmhits="")'
        )
    )
    # use dust filter?
    optional.add_argument(
        '--dust', metavar='yes/no', type=str,
        help='Use BLASTn dust filter during re-BLAST. Greatly decreases runtime if reference miRNA(s) '
             'are located in repeat regions. '
             'However ncOrtho will also not identify orthologs for these miRNAs',
        nargs='?',
        const='no', default='no'
    )
    # check Co-ortholog-ref
    optional.add_argument(
        '--checkCoorthologsRef', type=str2bool, metavar='True/False', nargs='?', const=False, default=False,
        help=(
            'If the re-blast does not identify the original reference miRNA sequence as best hit,'
            'ncOrtho will check whether the best blast '
            'hit is likely a co-ortholog of the reference miRNA relative to the search taxon. '
            'NOTE: Setting this flag will substantially increase'
            'the sensitivity of HaMStR but most likely affect also the specificity, '
            'especially when the search taxon is evolutionarily only very'
            'distantly related to the reference taxon (Default: False)'
        )
    )
    ##########################################################################################
    # Parse arguments
    ##########################################################################################

    # Show help when no arguments are added.
    if len(sys.argv) == 1:
        parser.print_help()
        sys.exit(1)
    else:
        args = parser.parse_args()
    
    # Check if computer provides the desired number of cores.
    available_cpu = mp.cpu_count()
    if args.cpu > available_cpu:
        print(
            '# Error: The provided number of CPU cores is higher than the '
            'number available on this system. Exiting...'
        )
        sys.exit(1)
    else:
        cpu = args.cpu

    # mandatory
    mirnas = args.ncrna
    models = args.models
    output = args.output
    query = args.query
    reference = args.reference

    # optional
    max_hits = args.maxcmhits
    cm_cutoff = args.cm_cutoff
    checkCoorthref = args.checkCoorthologsRef
    cleanup = args.cleanup
    heuristic = (args.heuristic, args.heur_blast_evalue, args.heur_blast_length)
    msl = args.minlength
    refblast = args.refblast
    qblast = args.queryblast
    dust = args.dust.strip()

    ##########################################################################################
    # Starting argument checks
    ##########################################################################################

    # Test if optional query name was given
    if args.queryname:
        qname = args.queryname
    else:
        qname = '.'.join(query.split('/')[-1].split('.')[:-1])
    # make folder for query in output dir
    if not output.split('/')[-1] == qname:
        output = f'{output}/{qname}'

    # Test if input files exist
    all_files = [mirnas, reference, query]
    for pth in all_files:
        if not os.path.isfile(pth):
            print(f'ERROR: {pth} is not a file')
            sys.exit()
    if not os.path.isdir(models):
        print(f'ERROR: Directory with covariance models does not exist at: {models}')
        sys.exit()

    # create symbolic links to guarantee writing permissions for pyfaidx index
    q_data = '{}/data'.format(output)
    if not os.path.isdir(q_data):
        os.makedirs(q_data)
    qlink = f'{q_data}/{qname}.fa'
    try:
        os.symlink(query, qlink)
    except FileExistsError:
        pass

    # check if reference BLASTdb was given as input
    if refblast:
        # check if BLASTdb exists
        if not check_blastdb(refblast):
            print('# Reference BLASTdb not found at: {}'.format(refblast))
            sys.exit()
    else:
        # check if refblast exists at location of reference genome
        if not check_blastdb(reference):
            refname = reference.split('/')[-1]
            refblast = f'{q_data}/{refname}'
            # check if BLASTdb exists already. Otherwise create it
            if not check_blastdb(refblast):
                make_blastndb(reference, refblast)
        else:
            refblast = reference
    # check if query BLASTdb was given as input (only used in heuristic mode)
    if qblast and heuristic[0]:
        # check if qblast exists
        if not check_blastdb(qblast):
            print('# Query BLASTdb not found at: {}'.format(qblast))
            sys.exit()
    elif heuristic[0]:
        qblast = qlink
        # check if already exists
        if not check_blastdb(qlink):
            print('# Creating query Database')
            make_blastndb(query, qlink)

    ###############################################################################################
    # Main
    ###############################################################################################

    # Print out query
    print('### Starting ncOrtho run for {}\n'.format(qname))

    # Create miRNA objects from the list of input miRNAs.
    mirna_dict = mirna_maker(mirnas, models, output, msl)

    outpath = '{0}/{1}_orthologs.fa'.format(output, qname)
    log_file = f'{output}/{qname}.log'
    with open(log_file, 'w') as log, open(outpath, 'w') as of:
        log.write(f'# miRNA\tStatus\n')
        # Identify ortholog candidates.
        for mirna in mirna_dict:
            restricted = False
            sys.stdout.flush()
            mirna_data = mirna_dict[mirna]
            # mirna objects for which no CM was found are empty
            if not mirna_data:
                log.write(f'{mirna}\tNo CM\n')
                continue
            print(f'\n### {mirna}')

            # Create output folder, if not existent.
            if not heuristic[0] or not cleanup:
                outdir = '{}/{}'.format(output, mirna)
            else:
                outdir = '{}'.format(output)
            if not os.path.isdir(outdir):
                os.makedirs(outdir)

            # start cmsearch
            cm_results, exitstatus = cmsearcher(
                mirna_data, cm_cutoff, cpu, msl, models, qlink, qblast, outdir, cleanup, heuristic
            )

            # Extract sequences for candidate hits (if any were found).
            if not cm_results:
                print('# {} for {}'.format(exitstatus, mirna))
                log.write(f'{mirna}\t{exitstatus}\n')
                continue
            elif max_hits and len(cm_results) > max_hits:
                print('# Maximum CMsearch hits reached. Restricting to best {} hits'.format(max_hits))
                cm_results = {k: cm_results[k] for k in list(cm_results.keys())[:max_hits]}
                restricted = True
            # get sequences for each CM result
            gp = GenomeParser(qlink, cm_results.values())
            candidates = gp.extract_sequences()
            nr_candidates = len(candidates)
            if nr_candidates == 1:
                print(
                    '# Covariance model search successful, found 1 '
                    'ortholog candidate.'
                )
            else:
                print(
                    '# Covariance model search successful, found {} '
                    'ortholog candidates.'
                    .format(nr_candidates)
                )
            print('# Evaluating candidates.')

            # Perform reverse BLAST test to verify candidates, stored in
            # a list (accepted_hits).
            reblast_hits = {}
            for candidate in candidates:
                sequence = candidates[candidate]
                # print('# Starting reverse blast for {}'.format(candidate))
                blast_command = (
                    'blastn -task blastn -db {0} -num_threads {1} -dust {2} -outfmt "6 qseqid sseqid pident '
                    'length mismatch gapopen qstart qend sstart send evalue bitscore sseq"'
                        .format(refblast, cpu, dust)
                )
                blast_command = (
                    'blastn -task blastn -db {0} -num_threads {1} -dust {2} -outfmt "6 sseqid '
                    'sstart send sstrand"'
                    .format(refblast, cpu, dust)
                )
                blast_call = sp.Popen(
                    blast_command, shell=True, stdin=sp.PIPE,
                    stdout=sp.PIPE, stderr=sp.PIPE, encoding='utf8'
                )
                # run BLAST
                res, err = blast_call.communicate(sequence)
                if err:
                    print(f'ERROR: {err}')
                    continue
                if not cleanup:
                    blast_output = '{0}/reBLAST_{1}.out'.format(outdir, candidate)
                    with open(blast_output, 'w') as bh:
                        for line in res:
                            bh.write(line)
                # parse BLASTn results
                blast_output = [line.split() for line in res.split('\n')]
                if not blast_output[0]:
                    print('No re-BLAST hits')
                    continue

                # BlastParser will read the temp_fasta file
                bp = BlastParser(mirna_data, blast_output, msl)
                # the parse_blast_output function will return True if the candidate is accepted
                if bp.evaluate_besthit():
                    print('Found best hit')
                    reblast_hits[candidate] = sequence
                elif checkCoorthref:
                    print('Best hit differs from reference sequence! Doing further checks\n')
                    # coorth_out = '{0}/{1}_coorth'.format(outdir, candidate)
                    if bp.check_coortholog_ref(sequence, outdir):
                        reblast_hits[candidate] = sequence
                else:
                    print('Best hit does not overlap with miRNA location')

            # Write output file if at least one candidate got accepted.
            if reblast_hits:
                nr_accepted = len(reblast_hits)
                if nr_accepted == 1:
                    print('# ncOrtho found 1 verified ortholog.')
                    out_dict = reblast_hits
                else:
                    print(
                        '# ncOrtho found {} potential co-orthologs.'
                        .format(nr_accepted)
                    )
                    print('# Evaluating distance between candidates to verify co-orthologs')
                    rbp = ReBlastParser(mirna_data, reblast_hits)
                    out_dict = rbp.verify_coorthologs(outdir)
                    if len(out_dict) == 1:
                        print('ncOrtho found 1 verified ortholog')
                    else:
                        print(
                            '# ncOrtho found {} verified co-orthologs.'
                                .format(len(out_dict))
                        )

                for hit in out_dict:
                    cmres = list(cm_results[hit])
                    cmres.insert(0, qname)
                    cmres = [str(entry) for entry in cmres]
                    header = '|'.join(cmres)
                    of.write('>{0}\n{1}\n'.format(header, out_dict[hit]))

                log.write(f'{mirna}\tSUCESS\n')
            else:
                print(
                    '# None of the candidates for {} could be verified.'
                    .format(mirna)
                )
                if restricted:
                    log.write(f'{mirna}\tNo re-BLAST after restricting to {max_hits} CMsearch hits\n')
                else:
                    log.write(f'{mirna}\tNo re-BLAST\n')
        print('\n### ncOrtho is finished!')


if __name__ == "__main__":
    main()
